#
# Copyright (c) 2012-2014 Fixpoint Inc. All rights reserved.
# ---
#
#   セットアッププログラム用共通関数定義
#
#
shopt -s extglob
BRANCH=1.5
VERSION=1.5.5.post10
SHORT_VERSION=1.5.5


trap "interrupt" SIGHUP SIGINT SIGQUIT SIGABRT SIGTERM

REQUIRES_ARCH="x86_64"
REQUIRES_PYTHON="Python 2.[67].*"
REQUIRES_PYTHON26="Python 2.6.*"
REQUIRES_PYTHON27="Python 2.7.*"
REQUIRES_POSTGRES="9.6"
REQUIRES_SYSTEM_CENT="CentOS*release [67].*"
REQUIRES_SYSTEM_AMZN="Amazon Linux AMI release*"
REQUIRES_SYSTEM_RHEL="Red Hat Enterprise Linux Server release [67].*"

#
# システム種別チェック
#
: ${SYSTEM_RELEASE:=}
: ${RHEL_VERSION:=$(rpm -E '%{rhel}' 2>/dev/null)}
if [ -z "$SYSTEM_RELEASE" ]; then
    for file in /etc/system-release /etc/redhat-release; do
        if [ -f $file ]; then SYSTEM_RELEASE=$(< $file); break; fi
    done
fi
case $SYSTEM_RELEASE in
    $REQUIRES_SYSTEM_CENT) SYSTEM="CENT"; SYSTEM_NAME="cent$RHEL_VERSION" ;;
    $REQUIRES_SYSTEM_RHEL) SYSTEM="RHEL"; SYSTEM_NAME="rhel$RHEL_VERSION" ;;
    $REQUIRES_SYSTEM_AMZN) SYSTEM="AMZN"; SYSTEM_NAME="amzn" ;;
    *)                     SYSTEM="" ;;
esac
if [ -z "$SYSTEM_RELEASEVER" ]; then
    SYSTEM_RELEASEVER=$(echo "$SYSTEM_RELEASE" | sed -r -e "s/.*release ([0-9.]+).*/\1/")
fi

#
# [RHEL] RHUI チェック
#
if [ -z "$RHUI_MODE" ]; then
    if [ $SYSTEM == "RHEL" ] && grep -qiE "^\[rhui\b.*\]" /etc/yum.repos.d/*.repo; then
        RHUI_MODE=true
    else
        RHUI_MODE=false
    fi
fi

#
# systemd 判別チェック
#
if [ "$(cat /proc/1/comm 2>/dev/null)" = "systemd" ]; then
    SYSTEMD=true
else
    SYSTEMD=false
fi
SHOW_OPTIONS="SYSTEM SYSTEM_RELEASE SYSTEM_RELEASEVER PYTHON SYSTEMD TMPDIR"

: ${PYTHON:=$((which python27 || which python2.7 || which python) 2>/dev/null)}
: ${PIP:=$(which pip 2>/dev/null)}
: ${SERVICE:=service}
: ${CHKCONFIG:=chkconfig}
: ${SYSTEMCTL:=systemctl}
: ${INITCTL:=initctl}
: ${INSTALL:=install}
: ${SUBSCRIPTION_MANAGER:=$(which subscription-manager 2>/dev/null)}
: ${NETWORK_SCRIPTS_DIR:=/etc/sysconfig/network-scripts}
: ${TMPDIR_TEMPLATE:=".tmp.install-$(date +%Y%m%d-%H%M).XXXX"}
: ${TMPDIR:="/tmp"}
: ${SETUP_TYPE:="Setup"}
: ${SKIP_CLEAN_YUM_CACHE:=true}
: ${OFFLINE_MODE:=false}

#
# runuser/su チェック
#
if [ -z "$SU" ]; then
    if [ -x /sbin/runuser ]; then
        SU=/sbin/runuser
    else
        SU=su
    fi
fi

#
# variables for Kompira
#
: ${KOMPIRA_USER:=kompira}
: ${KOMPIRA_GROUP:=kompira}
: ${KOMPIRA_HOME:=/opt/kompira}
: ${KOMPIRA_VAR_DIR:=/var/opt/kompira}
: ${KOMPIRA_LOG_DIR:=/var/log/kompira}
: ${KOMPIRA_BACKUP:="$KOMPIRA_VAR_DIR/kompira_backup.json.gz"}
: ${KOMPIRA_ENV:="$KOMPIRA_HOME"}
: ${KOMPIRA_BIN:="$KOMPIRA_HOME/bin"}
: ${KOMPIRA_PG_USER:=kompira}
: ${KOMPIRA_PG_DATABASE:=kompira}
: ${KOMPIRA_REALM_NAME:=default}

#
# variables for extra package
#
: ${KOMPIRA_EXTRA_PREFIX:="kompira-extra"}
: ${KOMPIRA_EXTRA_SUFFIX:=""}
: ${KOMPIRA_EXTRA_NAME:="$KOMPIRA_EXTRA_PREFIX-$SHORT_VERSION.$SYSTEM_NAME$KOMPIRA_EXTRA_SUFFIX"}
: ${KOMPIRA_EXTRA_DIR:="$KOMPIRA_EXTRA_NAME"}
: ${KOMPIRA_EXTRA_REPO:="/etc/yum.repos.d/$KOMPIRA_EXTRA_NAME.repo"}
: ${KOMPIRA_EXTRA_BASE:="/opt/kompira/extra/$SHORT_VERSION/$SYSTEM_NAME"}
: ${KOMPIRA_EXTRA_PIP:="$KOMPIRA_EXTRA_BASE/pip"}
: ${KOMPIRA_EXTRA_RPM:="$KOMPIRA_EXTRA_BASE/packages"}
: ${KOMPIRA_EXTRA_WHL:="$KOMPIRA_EXTRA_BASE/wheelhouse"}

#
# variables for proxy settings
#
: ${PROXY_URL:="$http_proxy"}
: ${NO_PROXY:="${no_proxy:-localhost,127.0.0.1}"}
if id $KOMPIRA_USER > /dev/null 2>&1; then
    if [ "OK" == "$($SU - $KOMPIRA_USER -c "echo OK" 2>/dev/null)" ]; then
        PROXY_URL=$($SU - $KOMPIRA_USER -c "echo \$http_proxy")
        NO_PROXY=$($SU - $KOMPIRA_USER -c "echo \$no_proxy")
    fi
fi

#
# variables for postgres
#
PG_VERSION="$REQUIRES_POSTGRES"
PG_PREFIX="postgresql${PG_VERSION/./}"
if [ "$SYSTEM" == "AMZN" ]; then
    PG_SERVICE="postgresql${PG_VERSION/./}"
else
    PG_SERVICE="postgresql-${PG_VERSION}"
    PG_REPONAME="pgdg${PG_VERSION/./}"
    PG_REPOPKG0="pgdg-redhat"
    PG_REPOPKG="${PG_REPOPKG0}${PG_VERSION/./}"
fi
PG_DATADIR="/var/lib/pgsql/${PG_VERSION}/data"
PG_BASEDIR=$(dirname $PG_DATADIR)
PG_LOG_DIR="/var/log/postgresql"
PG_REPL_USER="repl"
PG_REPL_PASS=$PG_REPL_USER
PGSQL_USER="postgres"
PGSQL_DB="postgres"
PGSQL_OPTIONS=""

#
# variables for cluster
#
: ${CLUSTER_RESOURCE_PATH:=/usr/lib/ocf/resource.d}
CLUSTER_CONFIGURED=false
CLUSTER_RUNNING=false
CLUSTER_MASTER=false
CLUSTER_SLAVE=false

#
# 冗長構成の対向ホスト名
#
HA_LOCALHOST=ha-kompira-local
HA_OTHERHOST=ha-kompira-other
: ${HA_LOCALNAME:=}
: ${HA_OTHERNAME:=$HA_OTHERHOST}

#
# 冗長構成のリソース属性値
#
PGSQL_MASTER_SCORE=1000
PGSQL_SLAVE_SCORE=100
RABBITMQ_MASTER_SCORE=10
RABBITMQ_SLAVE_SCORE=5
NA_SCORE="#NA"


#
# vault-centos レポジトリ情報
#
VAULT_CENTOS_REPONAME="vault-centos*"


: ${ECHO_LEVEL:=3}
ECHO_OPTIONS=
ECHO_PREFIX=
ECHO_POSTFIX=

# cp/install コマンドでのデフォルトバックアップ接尾辞
export SIMPLE_BACKUP_SUFFIX=.old

_now()
{
    date "+%F %T"
}

_echo()
{
    local msg
    for msg in "$@"; do
        echo $ECHO_OPTIONS "[$(_now)] ${ECHO_PREFIX}${msg}${ECHO_POSTFIX}"
    done
}

echo_debug()
{
    ECHO_PREFIX="DEBUG: " _echo "$@"
}
echo_verbose()
{
    ECHO_PREFIX="VERBOSE: " _echo "$@"
}
echo_info()
{
    ECHO_PREFIX="INFO: " _echo "$@"
}
echo_warn()
{
    ECHO_PREFIX="WARN: " _echo "$@"
}
echo_error()
{
    ECHO_PREFIX="ERROR: " _echo "$@"
}
echo_always()
{
    ECHO_PREFIX="****: " _echo "$@"
}
echo_param()
{
    echo_info "$(printf "    %-24s = %s" "$1" "$2")"
}
echo_title()
{
    echo_always "----------------------------------------------------------------" "$@" ""
}

verbose_run()
{
    echo_verbose "run: $*" > /dev/stderr
    "$@"
    local rc=$?
    if [ "$rc" -ne 0 ]; then
        echo_verbose "status=$rc" > /dev/stderr
    fi
    return $rc
}

expand_template()
{
    eval "$(echo "cat <<__EOF__"; cat $1; echo "__EOF__")"
}

diff_install()
{
    diff_and_cmd $INSTALL "$@"
}

diff_cp()
{
    diff_and_cmd cp -b "$@"
}

diff_and_cmd()
{
    local old=${!#}
    set ${@:1:$#-1}
    local new=${!#}
    set ${@:1:$#-1}
    if ! diff -N -u "$old" "$new"; then
        verbose_run "$@" "$new" "$old"
    fi
}

interrupt()
{
    echo_title "Abort ${SETUP_TYPE}."
    exit 1
}

inquire()
{
    echo -n "$1 "
    if [ -t 0 ]; then
        read answer
        case $answer in
            [yY]* ) answer="y";;
            [nN]* ) answer="n";;
            *) answer="$2" ;;
        esac
    else
        answer="$2"
        abort_setup "stdin is not the terminal!"
    fi
}

install_repo_direct()
{
    local repo_name=$1
    local repo_file=$2
    local repo_url=$3

    if [ ! -f /etc/yum.repos.d/$repo_file ]
    then
        echo_info "Install repository (direct): $repo_name"
        verbose_run yum-config-manager --add-repo=$repo_url
        exit_if_failed "$?" "Failed to install $repo_name"
    fi
}

install_repo_from_rpm()
{
    local repo_name=$1
    local repo_remove=$2
    local repo_url=$3

    if ! is_yum_installed $repo_name
    then
        if [ -n "$repo_remove" ]; then
            remove_if_installed $repo_remove
        fi
        echo_info "Install repository (rpm): $repo_name"
        local dir=$(dirname $repo_url)
        local pattern=$(basename $repo_url)
        echo_verbose "search $pattern in $dir"
        local url=$(find_link $dir $pattern | tail -1)
        if [ -z "$url" ]; then
            exit_if_failed 1 "$repo_name is not found"
        fi
        echo_info "install $url"
        verbose_run yum $YUM_OPTION -y install $url
        exit_if_failed "$?" "Failed to install $repo_name"
        return 0
    fi
    return 1
}

install_vault_repo()
{
    echo_info "Install vault.centos.org repository"
    if [ -z "$CENT_RELEASEVER" ]; then
        #
        # vault.centos.org に SYSTEM_RELEASEVER で始まるリンクが含まれるか確認する
        #
        CENT_RELEASEVER=$(curl -sq "http://vault.centos.org/" | grep -o -m1 -E "href=\"$SYSTEM_RELEASEVER(|\.[0-9.]+)/\"" | sed -r -e 's/.*"([0-9.]+).*/\1/')
        if [ -z "$CENT_RELEASEVER" ]; then
            abort_setup "Mirror repository for CentOS $SYSTEM_RELEASEVER not found in vault.centos.org"
        fi
    fi
    expand_template $THIS_DIR/config/cluster/vault-centos.repo > $TMPDIR/vault-centos.repo
    # diff_cp $TMPDIR/vault-centos.repo /etc/yum.repos.d/vault-centos.repo
    verbose_run yum-config-manager --add-repo=$TMPDIR/vault-centos.repo
    exit_if_failed "$?" "Failed to install vault-centos.repo"
}

find_link()
{
    #
    # url で指定した HTML ページから pattern にマッチするリンク(href)を抽出する
    #
    local url=$1
    local pattern=$2
    curl -L -s "$url/" | sed -nre "s|.*href=\"($pattern)\".*|$url/\1|p"
}

is_runnable()
{
    test -x "$(which $1 2>/dev/null)"
}

is_yum_installed()
{
    local pkg
    local yum_option=$YUM_OPTION
    if $OFFLINE_MODE; then
        yum_option="$yum_option --noplugins --disablerepo=* --enablerepo=$KOMPIRA_EXTRA_NAME"
    fi
    for pkg; do
        if ! yum $yum_option list installed $pkg > /dev/null 2>&1
        then
            return 1
        fi
    done
}

exit_if_failed()
{
    local rc=$1
    shift
    if [ $rc -ne 0 ]; then
        abort_setup "$@"
    fi
}

yum_install()
{
    if [ -n "$*" ]; then
        local DESCRIPTION="${DESCRIPTION:-$@}"
        local yum_option=$YUM_OPTION
        echo_info "Install $DESCRIPTION"
        if $OFFLINE_MODE; then
            yum_option="$yum_option --noplugins --disablerepo=* --enablerepo=$KOMPIRA_EXTRA_NAME"
        fi
        verbose_run yum $yum_option -y install "$@"
        exit_if_failed "$?" "Failed to install $DESCRIPTION"
    fi
}

yum_install_if_not_installed()
{
    local DESCRIPTION="${DESCRIPTION:-$@}"
    if ! is_yum_installed $@; then
        yum_install $@
    fi
}

debuginfo_install()
{
    if [ -n "$*" ]; then
        local DESCRIPTION="${DESCRIPTION:-$@}"
        local yum_option=$YUM_OPTION
        echo_info "Debuginfo-install $DESCRIPTION"
        if $OFFLINE_MODE; then
            yum_option="$yum_option --noplugins --disablerepo=* --enablerepo=$KOMPIRA_EXTRA_NAME"
        fi
        verbose_run debuginfo-install $yum_option -y "$@"
        exit_if_failed "$?" "Failed to debuginfo-install $DESCRIPTION"
    fi
}

remove_if_installed()
{
    local pkg
    for pkg; do
        if is_yum_installed $pkg; then
            echo_info "Remove ${pkg}"
            verbose_run yum $YUM_OPTION -y remove $pkg
        fi
    done
}

ver2num()
{
    echo $1 | awk -F. '{printf "%03d%03d%03d%03d\n",$1,$2,$3,$4}'
}

get_rpm_version()
{
    local name=$1
    rpm -q "$name" --info | sed -rne 's/^Version\s*:\s*(\S+).*/\1/p'
}

get_pip_info()
{
    local name=$1
    local field=$2
    $PIP show "$name" 2> /dev/null | sed -rne "s/^$field:\s*(.*)/\1/p"
}

get_kompira_version()
{
    get_pip_info "Kompira" "Version"
}

get_kompira_location()
{
    get_pip_info "Kompira" "Location"
}

pip_uninstall_if_installed()
{
    local pkg
    for pkg; do
        if $PIP show $pkg > /dev/null; then
            echo_info "Uninstall ${pkg} [$PIP]"
            verbose_run $PIP uninstall -y $pkg
        fi
    done
}

is_activce_service()
{
    local svc=$1
    if $SYSTEMD; then
        $SYSTEMCTL is-active $svc > /dev/null 2>&1
    elif is_upstart_service $svc; then
        $INITCTL status $svc | grep -q "start/running"
    else
        $SERVICE $svc status > /dev/null 2>&1
    fi
}

is_upstart_service()
{
    local svc=$1
    if $SYSTEMD; then return 1; fi
    initctl status $svc > /dev/null 2>&1
}

_service_status()
{
    local svc=$1
    if $SYSTEMD; then
        verbose_run $SYSTEMCTL status $svc
    elif is_upstart_service $svc; then
        verbose_run $INITCTL status $svc
    else
        verbose_run $SERVICE $svc status
    fi
}

_service_start()
{
    local svc=$1
    if $SYSTEMD; then
        verbose_run $SYSTEMCTL start $svc
    elif is_upstart_service $svc; then
        verbose_run $INITCTL start $svc
    else
        verbose_run $SERVICE $svc start
    fi
}

_service_stop()
{
    local svc=$1
    if $SYSTEMD; then
        verbose_run $SYSTEMCTL stop $svc
    elif is_upstart_service $svc; then
        verbose_run $INITCTL stop $svc
    else
        verbose_run $SERVICE $svc stop
    fi
}

_service_restart()
{
    local svc=$1
    if $SYSTEMD; then
        verbose_run $SYSTEMCTL restart $svc
    elif is_upstart_service $svc; then
        verbose_run $INITCTL stop $svc
        verbose_run $INITCTL start $svc
    else
        verbose_run $SERVICE $svc restart
    fi
}

service_start()
{
    local svc=$1
    echo_info "Start service: ${svc}"
    _service_start $svc
    exit_if_failed "$?" "Failed to start ${svc}"
}

service_stop()
{
    local svc=$1
    if is_activce_service $svc; then
        echo_info "Stop service: ${svc}"
        _service_stop $svc
        exit_if_failed "$?" "Failed to stop ${svc}"
    fi
}

service_restart()
{
    local svc=$1
    if is_activce_service $svc; then
        echo_info "Restart service: ${svc}"
        _service_restart $svc
        exit_if_failed "$?" "Failed to restart ${svc}"
    else
        service_start $svc
    fi
}

service_register()
{
    local svc=$1
    if $SYSTEMD; then
        verbose_run $SYSTEMCTL daemon-reload
    else
        verbose_run $CHKCONFIG --add $svc
    fi
}

service_enable()
{
    local svc=$1
    if $SYSTEMD; then
        verbose_run $SYSTEMCTL enable $svc.service
    else
        verbose_run $CHKCONFIG $svc on
    fi
}

service_disable()
{
    local svc=$1
    if $SYSTEMD; then
        verbose_run $SYSTEMCTL disable $svc.service
    else
        verbose_run $CHKCONFIG $svc off
    fi
}

check_root()
{
    if [ $(whoami) != root ]; then
        abort_setup "Please execute as root user."
    fi
}

check_arch()
{
    local arch=$(uname -i)
    case $arch in
        $REQUIRES_ARCH)
            echo_info "Confirmed architecture: $arch" ;;
        *)
            abort_setup "Please execute on the x86_64 Linux." ;;
    esac
}

check_system()
{
    if [ -z "$SYSTEM" ]; then
        abort_setup "CentOS/Red Hat 6.X or 7.X is required!"
    fi
    echo_info "Confirmed system: $SYSTEM_RELEASE"
}

get_cluster_status()
{
    if ! is_yum_installed "pacemaker"; then
        echo_info "Pacemaker is not installed."
        return 0
    fi
    if [ -f $CLUSTER_RESOURCE_PATH/kompira/kompirad ]; then
        echo_info "Resource is configured."
        CLUSTER_CONFIGURED=true
    fi
    if ! is_activce_service "pacemaker"; then
        echo_info "Pacemaker is stop."
        return 0
    fi
    CLUSTER_RUNNING=true
    if pgsql_is_primary; then
        echo_info "Pacemaker is running (MASTER)"
        CLUSTER_MASTER=true
    else
        echo_info "Pacemaker is running (SLAVE)"
        CLUSTER_SLAVE=true
    fi
}

check_subscription()
{
    local offline_mode=${1:-false}
    if $offline_mode || $RHUI_MODE; then
        return 0
    fi
    if [[ $SYSTEM == "RHEL" && -n "$SUBSCRIPTION_MANAGER" ]]; then
        echo_info "[RHEL] Check subscription status"
        if ! $SUBSCRIPTION_MANAGER status; then
            abort_setup "Please attach a valid subscription." \
                        "" \
                        "ex) subscription-manager register --auto-attach"
        fi
    fi
}

pgsql_cmddir()
{
    dirname $(readlink -f $(which psql))
}

pgsql_replica()
{
    local master_name="$1"

    verbose_run rm -f $PG_BASEDIR/tmp/PGSQL.lock
    verbose_run rm -rf $PG_DATADIR.old
    verbose_run mv -f $PG_DATADIR $PG_DATADIR.old

    # pg_baseback コマンドでデータベースをバックアップする
    # postgres ユーザで実行するため、ルートディレクトリに移動してから起動する
    (cd / && verbose_run sudo -u postgres pg_basebackup -h $master_name -U $PG_REPL_USER -D $PG_DATADIR -X stream -P)
    local rc=$?
    if [ $rc -eq 0 ]; then
        # 成功した場合は $PG_DATADIR.old を削除する
        echo_info "Backup succeed: cleanup $PG_DATADIR.old"
        rm -rf $PG_DATADIR.old
    else
        # 失敗した場合は $PG_DATADIR.old を元に戻す
        echo_error "Backup FAILED!: restore $PG_DATADIR"
        rm -rf $PG_DATADIR
        mv -f $PG_DATADIR.old $PG_DATADIR
    fi
    return $rc
}

pgsql_has_kompira_table()
{
    # postgres が初期化されているか（kompira データベースにテーブルが存在するか）確認する
    LANG= psql -U kompira -Atc '\dt' kompira | grep -q "table"
}

pgsql_is_in_recovery()
{
    # postgres がリカバリモードか確認する
    psql -U $PGSQL_USER -Atc "select pg_is_in_recovery()" 2>/dev/null
}

pgsql_is_primary()
{
    local pg_is_in_recovery rc
    pg_is_in_recovery=$(pgsql_is_in_recovery)
    rc=$?
    if [ $rc -ne 0 ]; then
        echo_warn "PostgreSQL is stop (rc=$rc)"
        return 2
    fi
    case $pg_is_in_recovery in
        f)
            echo_info "PostgreSQL is running as a primary."
            return 0
            ;;
        t)
            echo_info "PostgreSQL is running as a hot standby."
            return 1
            ;;
        *)
            abort_setup "Invalid result: pg_is_in_recovery=$pg_is_in_recovery"
            ;;
    esac
}

setup_firewalld()
{
    echo_title "Configure firewalld settings."
    local rule rule_elem rule_spec prot port
    for rule; do
        case $rule in
            +([0-9])/@(tcp|udp))
                verbose_run firewall-cmd --add-port=$rule --permanent
                ;;
            +([0-9]|\.)?(/+([0-9])))
                verbose_run firewall-cmd --zone=trusted --add-source=$rule --permanent
                ;;
            *)
                verbose_run firewall-cmd --add-service=$rule --permanent
                ;;
        esac
    done
    verbose_run firewall-cmd --reload
}

setup_iptables()
{
    echo_title "Configure iptables settings."
    local rule rule_elem rule_spec prot port
    iptables-save > $TMPDIR/iptables_cur
    for rule; do
        rule_spec=""
        for rule_elem in ${rule/,/ }; do
            case $rule_elem in
                +([0-9])/@(tcp|udp))
                    prot=${rule_elem#*/}
                    port=${rule_elem%/*}
                    rule_spec="$rule_spec -p $prot -m $prot --dport $port"
                    ;;
                +([0-9]|\.)?(/+([0-9])))
                    rule_spec="$rule_spec -s $rule_elem"
                    ;;
                *)
                    echo_warn "invalid rule: $rule_elem"
                    ;;
             esac
        done
        rule_spec=${rule_spec# }
        if [ -z "$rule_spec" ]; then
            continue
        fi
        rule_spec="INPUT $rule_spec -j ACCEPT"
        if grep -q -e "^-A $rule_spec" $TMPDIR/iptables_cur; then
            echo_info "A rule '$rule_spec' is already set."
        else
            echo_info "Append a rule '$rule_spec' to iptables."
            verbose_run iptables -I $rule_spec
        fi
    done
    iptables-save > $TMPDIR/iptables_new
    diff_cp $TMPDIR/iptables_new /etc/sysconfig/iptables
}

cluster_wait_current_dc()
{
    local timeout=60
    ECHO_OPTIONS="-n" echo_info "Waiting Current DC"
    sleep 3
    while [ $timeout -gt 0 ]; do
        let timeout=$timeout-1
        if crm_mon -1 | grep "Current DC: NONE" > /dev/null; then
            echo -n "."
            sleep 1
            continue
        fi
        echo
        return 0
    done
    echo
    return 1
}

pcs_enter_maintenance_mode()
{
    echo_info "Enter the pacemaker in maintenance mode."
    verbose_run pcs property set maintenance-mode=true
}

pcs_leave_maintenance_mode()
{
    echo_info "Exit the pacemaker from maintenance mode."
    verbose_run pcs property set maintenance-mode=false
}

pcs_remove_resources()
{
    local pcs_opt=$@
    #
    # 既存のリソースを削除する
    #
    local res
    pcs_remove_constraints $pcs_opt
    for res in webserver ms_pgsql ms_rabbitmq; do
        verbose_run pcs $pcs_opt resource delete $res
    done
}

pcs_remove_operations()
{
    local pcs_opt=$@
    #
    # 既存のリソース動作を削除する
    #
    local operation
    for operation in $(pcs resource show --full | grep -E '(promote|demote|start|stop|monitor|notify).*\(.+\)' | sed -re 's/.*\((.*)\).*/\1/'); do
        verbose_run pcs $pcs_opt resource op remove $operation
    done
}

pcs_remove_constraints()
{
    local pcs_opt=$@
    #
    # 既存の制約条件を削除する
    #
    local constraint
    for constraint in $(pcs constraint show --full | grep 'id:' | sed -re 's/.*id:(.*)\)\s*$/\1/'); do
        verbose_run pcs $pcs_opt constraint remove $constraint
    done
}

pcs_setup_constraints()
{
    local pcs_opt=$@
    # pgsql が promote してから webserver を起動する
    # kompirad が早く起動して master になっていない pgsql に接続しようとするのを防ぐ
    verbose_run pcs $pcs_opt constraint order promote ms_pgsql then start webserver

    # webmaster を rabbitmq と pgsql の master セットにコロケートする
    # rabbitmq と pgsql が master となるノードで、webmaster は起動できる
    verbose_run pcs $pcs_opt constraint colocation set webserver set ms_rabbitmq ms_pgsql sequential=false role=Master

    # master pgsql を master rabbitmq にコロケートする
    # rabbitmq が master となるノードで、pgsql が master となることが許可される
    # MEMO: 本来は上記の colocation set 制約があれば不要と思われるが、
    #       rabbitmq が停止したときに pgsql が連動してフェイルオーバしないための対策
    verbose_run pcs $pcs_opt constraint colocation add master ms_pgsql with master ms_rabbitmq INFINITY
}

pcs_failcount_show()
{
    local res
    for res; do
        # [v1.5.0] pcs コマンドでは適切に failcount を扱えないので crm_failcount を利用する
        # pcs resource failcount show $res
        crm_failcount -r $res -G
    done
}

pcs_failcount_reset()
{
    local res
    for res; do
        # [v1.5.0] pcs コマンドでは適切に failcount を扱えないので crm_failcount を利用する
        # pcs resource failcount reset $res
        verbose_run crm_failcount -r $res -D
    done
}

pcs_show_status()
{
    echo_info "Display failcount for each resources."
    pcs_failcount_show res_vip res_httpd res_kompirad res_kompira_jobmngrd res_lsyncd res_pgsql res_rabbitmq
    echo_info "Display constraint setings of resources."
    pcs constraint show
    echo_info "Display state of resources."
    pcs resource show
}

check_localname()
{
    if [ -z $HA_LOCALNAME ]; then
        HA_LOCALNAME=$(crm_node -n)
        if [ -z $HA_LOCALNAME ]; then
            abort_setup "Name of local node could not be determined."
        fi
    fi
}

check_othername()
{
    if [ -z $HA_OTHERNAME ]; then
        HA_OTHERNAME=$(crm_node -l | cut -d' ' -f2 | grep -v -w $(crm_node -n) | head -n 1)
        if [ -z $HA_OTHERNAME ]; then
            abort_setup "Name of other node could not be determined."
        fi
    fi
}

get_rabbitmq_master_score()
{
    check_localname
    crm_attribute -l reboot -N "$HA_LOCALNAME" -n master-res_rabbitmq -G -q 2>/dev/null
}

get_pgsql_master_score()
{
    check_localname
    crm_attribute -l reboot -N "$HA_LOCALNAME" -n master-res_pgsql -G -q 2>/dev/null
}

get_pgsql_data_status()
{
    check_localname
    crm_attribute -l forever -N "$HA_LOCALNAME" -n res_pgsql-data-status -G -q 2>/dev/null
}

get_pgsql_status()
{
    check_localname
    crm_attribute -l reboot -N "$HA_LOCALNAME" -n res_pgsql-status -G -q 2>/dev/null
}

get_pgsql_attrset()
{
    echo "$(get_pgsql_status)/$(get_pgsql_data_status)/$(get_pgsql_master_score)"
}

wait_resources_stabilize()
{
    check_localname
    local pgsql_waiting=${1:-true}
    local rabbitmq_waiting=${2:-true}
    local pgsql_stabilized=false
    local rabbitmq_stabilized=false
    local timeout=60
    local stable=10
    local description
    local pgsql_curr pgsql_last pgsql_status pgsql_data_status pgsql_master_score
    local pgsql_count=0
    local pgsql_master="PRI/LATEST/$PGSQL_MASTER_SCORE"
    local pgsql_slave="HS:async/STREAMING|ASYNC/$PGSQL_SLAVE_SCORE"
    local pgsql_pending="HS:alone/LATEST/-INFINITY"
    local rabbitmq_curr rabbitmq_last rabbitmq_master_score
    local rabbitmq_master="$RABBITMQ_MASTER_SCORE"
    local rabbitmq_slave="$RABBITMQ_SLAVE_SCORE"
    local rabbitmq_count=0
    local header="    RabbitMQ  | PostgreSQL                          | description"
    local format="    %-10s| %-36s|%s"

    echo_info "Waiting for the resources to stabilize."
    echo "$header"
    while [ $timeout -gt 0 ]; do
        let timeout=$timeout-1
        #
        # リソースの状態を取得する
        #
        pgsql_status=$(get_pgsql_status)
        pgsql_data_status=$(get_pgsql_data_status)
        pgsql_master_score=$(get_pgsql_master_score)
        if [ -z $pgsql_master_score ]; then
            pgsql_master_score=$NA_SCORE
        fi
        pgsql_curr="$pgsql_status/$pgsql_data_status/$pgsql_master_score"
        rabbitmq_master_score=$(get_rabbitmq_master_score)
        if [ -z $rabbitmq_master_score ]; then
            rabbitmq_master_score=$NA_SCORE
        fi
        rabbitmq_curr="$rabbitmq_master_score"
        description=""
        #
        # pgsql の状態をチェックする
        #
        if $pgsql_waiting; then
            if [[ $pgsql_curr != $pgsql_pending && $pgsql_curr == $pgsql_last ]]; then
                let pgsql_count=$pgsql_count+1
            else
                pgsql_count=0
                pgsql_last=$pgsql_curr
            fi
            if [ $pgsql_curr == $pgsql_master ]; then
                pgsql_stabilized=true
                pgsql_waiting=false
                description="$description PostgreSQL has promoted as a Master."
            elif [ $pgsql_curr == $pgsql_slave ]; then
                pgsql_stabilized=true
                pgsql_waiting=false
                description="$description PostgreSQL has demoted as a Slave."
            elif [ $pgsql_count -ge $stable ]; then
                if [ $pgsql_status == "STOP" ]; then
                    description="$description [WARN] PostgreSQL is stopped."
                else
                    description="$description [WARN] PostgreSQL shows an unexpected state."
                fi
                pgsql_waiting=false
            fi
        fi
        #
        # rabbitmq の状態をチェックする
        #
        if $rabbitmq_waiting; then
            if [[ $rabbitmq_curr == $rabbitmq_last ]]; then
                let rabbitmq_count=$rabbitmq_count+1
            else
                rabbitmq_count=0
                rabbitmq_last=$rabbitmq_curr
            fi
            if [ $rabbitmq_curr == $rabbitmq_master ]; then
                rabbitmq_stabilized=true
                rabbitmq_waiting=false
                description="$description RabbitMQ as promoted as a Master."
            elif [ $rabbitmq_curr == $rabbitmq_slave ]; then
                rabbitmq_stabilized=true
                rabbitmq_waiting=false
                description="$description RabbitMQ as demoted as a Slave."
            elif [ $rabbitmq_count -ge $stable ]; then
                description="$description [WARN] RabbitMQ shows an unexpected state."
                rabbitmq_waiting=false
            fi
        fi
        #
        # 状態を表示する
        #
        printf "$format\n" "$rabbitmq_curr" "$pgsql_curr" "$description"
        #
        # 対象リソースが安定したら終了する
        #
        if $pgsql_stabilized && $rabbitmq_stabilized; then
            echo_info "Resources stablized."
            return 0
        elif ! $pgsql_waiting && ! $rabbitmq_waiting; then
            break
        fi
        sleep 1
    done
    return 1
}

normalize_proxy_url()
{
    local url="$1"
    case "$url" in
        "" | http*://) url="";;
        http*://*) ;;
        *) url="http://$url" ;;
    esac
    echo $url
}

urlparse()
{
    local URL="$1"
    local URLPARSE_PREFIX="${URLPARSE_PREFIX:-URLPARSE_}"
    local URLPARSE_KEYS="${URLPARSE_KEYS:-scheme,hostname,port,username,password}"
    $PYTHON -c "from urlparse import urlparse; url=urlparse('$URL'); print '\n'.join(['${URLPARSE_PREFIX}%s=%s' % (k.upper(), getattr(url, k) or '') for k in '$URLPARSE_KEYS'.split(',')])"
}

show_options()
{
    for opt in $@; do
        echo_param "$opt" "${!opt}"
    done
}

create_tmpdir()
{
    TMPDIR=$(mktemp -d --tmpdir=$(pwd) "$TMPDIR_TEMPLATE")
}

start_setup()
{
    echo_always \
        "****************************************************************" \
        "Kompira-$VERSION:" \
        "Start: ${SETUP_TYPE}" \
        ""
    show_options $SHOW_OPTIONS
}

exit_setup()
{
    local rc=${1:-0}
    echo_always \
        "" \
        "Finish: ${SETUP_TYPE} (status=$rc)" \
        "****************************************************************"
    sleep 1
    exit $rc
}

abort_setup()
{
    echo_error "" "$@" ""
    exit_setup 1
}
